@@ -1,23 +1,27 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import hashlib
+
+from collections import defaultdict
 from decimal import Decimal
+from itertools import groupby
 
 from sql import Null
 from sql.aggregate import Sum
 from sql.conditionals import Case, Coalesce
 from sql.functions import Abs
 
-from trytond import backend
 from trytond.i18n import gettext
-from trytond.pool import Pool, PoolMeta
 from trytond.model import ModelSQL, ModelView, fields
-from trytond.pyson import Eval, If, Bool, Id
-from trytond.wizard import (Wizard, StateView, StateAction, StateTransition,
-    Button)
-from trytond.transaction import Transaction
-from trytond.tools.multivalue import migrate_property
+from trytond.modules.account.exceptions import (
+    CancelWarning, DelegateLineWarning, GroupLineWarning,
+    RescheduleLineWarning)
 from trytond.modules.company.model import CompanyValueMixin
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, Id, If
+from trytond.transaction import Transaction, check_access
+from trytond.wizard import (
+    Button, StateAction, StateTransition, StateView, Wizard)
 
 from .exceptions import BlockedWarning, GroupWarning
 from .payment import KINDS
@@ -25,17 +29,13 @@ from .payment import KINDS
 
 class MoveLine(metaclass=PoolMeta):
     __name__ = 'account.move.line'
-    payment_amount = fields.Function(fields.Numeric('Payment Amount',
-            digits=(16,
-                If(Bool(Eval('second_currency_digits')),
-                    Eval('second_currency_digits', 2),
-                    Eval('currency_digits', 2))),
+    payment_amount = fields.Function(Monetary(
+            "Amount to Pay",
+            currency='payment_currency', digits='payment_currency',
             states={
                 'invisible': ~Eval('payment_kind'),
-                },
-            depends=['payment_kind', 'second_currency_digits',
-                'currency_digits']), 'get_payment_amount',
-        searcher='search_payment_amount')
+                }),
+        'get_payment_amount', searcher='search_payment_amount')
     payment_currency = fields.Function(fields.Many2One(
             'currency.currency', "Payment Currency"),
         'get_payment_currency')
@@ -43,8 +43,7 @@ class MoveLine(metaclass=PoolMeta):
         readonly=True,
         states={
             'invisible': ~Eval('payment_kind'),
-            },
-        depends=['payment_kind'])
+            })
     payment_kind = fields.Function(fields.Selection([
                 (None, ''),
                 ] + KINDS, 'Payment Kind'), 'get_payment_kind')
@@ -55,7 +54,6 @@ class MoveLine(metaclass=PoolMeta):
                 (Eval('payment_kind') == 'payable')
                 & ((Eval('credit', 0) > 0) | (Eval('debit', 0) < 0))),
             },
-        depends=['payment_kind', 'debit', 'credit'],
         help="Check if the line will be paid by direct debit.")
 
     @classmethod
@@ -63,16 +61,23 @@ class MoveLine(metaclass=PoolMeta):
         super(MoveLine, cls).__setup__()
         cls._buttons.update({
                 'pay': {
-                    'invisible': ~Eval('payment_kind').in_(
-                        list(dict(KINDS).keys())),
+                    'invisible': (
+                        ~Eval('payment_kind').in_(list(dict(KINDS).keys()))
+                        | Eval('reconciliation')),
                     'depends': ['payment_kind'],
                     },
                 'payment_block': {
-                    'invisible': Eval('payment_blocked', False),
+                    'invisible': (
+                        ~Eval('payment_kind').in_(list(dict(KINDS).keys()))
+                        | Eval('reconciliation')
+                        | Eval('payment_blocked', False)),
                     'depends': ['payment_blocked'],
                     },
                 'payment_unblock': {
-                    'invisible': ~Eval('payment_blocked', False),
+                    'invisible': (
+                        ~Eval('payment_kind').in_(list(dict(KINDS).keys()))
+                        | Eval('reconciliation')
+                        | ~Eval('payment_blocked', False)),
                     'depends': ['payment_blocked'],
                     },
                 })
@@ -188,6 +193,83 @@ class MoveLine(metaclass=PoolMeta):
                 'payment_blocked': False,
                 })
 
+    @classmethod
+    def _pay_direct_debit_domain(cls, date):
+        return [
+            ['OR',
+                ('account.type.receivable', '=', True),
+                ('account.type.payable', '=', True),
+                ],
+            ('party', '!=', None),
+            ('reconciliation', '=', None),
+            ('payment_amount', '!=', 0),
+            ('move_state', '=', 'posted'),
+            ['OR',
+                ('debit', '>', 0),
+                ('credit', '<', 0),
+                ],
+            ('maturity_date', '<=', date),
+            ('payment_blocked', '!=', True),
+            ]
+
+    @classmethod
+    def pay_direct_debit(cls, date=None):
+        pool = Pool()
+        Date = pool.get('ir.date')
+        Payment = pool.get('account.payment')
+        Reception = pool.get('party.party.reception_direct_debit')
+        if date is None:
+            date = Date.today()
+        with check_access():
+            lines = cls.search(cls._pay_direct_debit_domain(date))
+
+        payments = []
+        for line in lines:
+            if not line.payment_amount:
+                # SQLite fails to search for payment_amount != 0
+                continue
+            pattern = Reception.get_pattern(line)
+            for reception in line.party.reception_direct_debits:
+                if reception.match(pattern):
+                    payments.extend(reception.get_payments(line))
+                    break
+        Payment.save(payments)
+        return payments
+
+
+class CreateDirectDebit(Wizard):
+    "Create Direct Debit"
+    __name__ = 'account.move.line.create_direct_debit'
+
+    start = StateView('account.move.line.create_direct_debit.start',
+        'account_payment.move_line_create_direct_debit_view_form', [
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Create", 'create_', 'tryton-ok', default=True),
+            ])
+    create_ = StateAction('account_payment.act_payment_form')
+
+    def do_create_(self, action):
+        pool = Pool()
+        Line = pool.get('account.move.line')
+        payments = Line.pay_direct_debit(date=self.start.date)
+        action['domains'] = []
+        return action, {
+            'res_id': [p.id for p in payments],
+            }
+
+
+class CreateDirectDebitStart(ModelView):
+    "Create Direct Debit"
+    __name__ = 'account.move.line.create_direct_debit.start'
+
+    date = fields.Date(
+        "Date", required=True,
+        help="Create direct debit for lines due up to this date.")
+
+    @classmethod
+    def default_date(cls):
+        return Pool().get('ir.date').today()
+
 
 class PayLineStart(ModelView):
     "Pay Line"
@@ -207,8 +289,7 @@ class PayLineAskJournal(ModelView):
         required=True, domain=[
             ('company', '=', Eval('company', -1)),
             ('currency', '=', Eval('currency', -1)),
-            ],
-        depends=['company', 'currency'])
+            ])
 
     journals = fields.One2Many(
         'account.payment.journal', None, 'Journals', readonly=True)
@@ -268,9 +349,8 @@ class PayLine(Wizard):
                         continue
                     lines = [l for l in types[kind]['lines']
                         if l.party == party]
-                    warning_name = '%s:%s:%s' % (
-                        reverse[kind], party,
-                        hashlib.md5(str(lines).encode('utf-8')).hexdigest())
+                    warning_name = Warning.format(
+                        '%s:%s' % (reverse[kind], party), lines)
                     if Warning.check(warning_name):
                         names = ', '.join(l.rec_name for l in lines[:5])
                         if len(lines) > 5:
@@ -351,10 +431,7 @@ class PayLine(Wizard):
             amount=line.payment_amount,
             line=line,
             )
-        date = self.start.date or line.maturity_date
-        # Use default value when empty
-        if date:
-            payment.date = date
+        payment.date = self.start.date or line.maturity_date
         return payment
 
     def do_pay(self, action):
@@ -407,26 +484,7 @@ class ConfigurationPaymentGroupSequence(ModelSQL, CompanyValueMixin):
             ('company', 'in', [Eval('company', -1), None]),
             ('sequence_type', '=',
                 Id('account_payment', 'sequence_type_account_payment_group')),
-            ],
-        depends=['company'])
-
-    @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-
-        super(ConfigurationPaymentGroupSequence, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('payment_group_sequence')
-        value_names.append('payment_group_sequence')
-        fields.append('company')
-        migrate_property(
-            'account.configuration', field_names, cls, value_names,
-            fields=fields)
+            ])
 
     @classmethod
     def default_payment_group_sequence(cls):
@@ -439,6 +497,104 @@ class ConfigurationPaymentGroupSequence(ModelSQL, CompanyValueMixin):
             return None
 
 
+class MoveCancel(metaclass=PoolMeta):
+    __name__ = 'account.move.cancel'
+
+    def transition_cancel(self):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        moves_w_payments = []
+        for move in self.records:
+            for line in move.lines:
+                if any(p.state != 'failed' for p in line.payments):
+                    moves_w_payments.append(move)
+                    break
+        if moves_w_payments:
+            names = ', '.join(
+                m.rec_name for m in moves_w_payments[:5])
+            if len(moves_w_payments) > 5:
+                names += '...'
+            key = Warning.format('cancel_payments', moves_w_payments)
+            if Warning.check(key):
+                raise CancelWarning(
+                    key, gettext(
+                        'account_payment.msg_move_cancel_payments',
+                        moves=names))
+        return super().transition_cancel()
+
+
+class MoveLineGroup(metaclass=PoolMeta):
+    __name__ = 'account.move.line.group'
+
+    def do_group(self, action):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        lines_w_payments = []
+        for line in self.records:
+            if any(p.state != 'failed' for p in line.payments):
+                lines_w_payments.append(line)
+        if lines_w_payments:
+            names = ', '.join(
+                m.rec_name for m in lines_w_payments[:5])
+            if len(lines_w_payments) > 5:
+                names += '...'
+            key = Warning.format('group_payments', lines_w_payments)
+            if Warning.check(key):
+                raise GroupLineWarning(
+                    key, gettext(
+                        'account_payment.msg_move_line_group_payments',
+                        lines=names))
+        return super().do_group(action)
+
+
+class MoveLineReschedule(metaclass=PoolMeta):
+    __name__ = 'account.move.line.reschedule'
+
+    def do_reschedule(self, action):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        lines_w_payments = []
+        for line in self.records:
+            if any(p.state != 'failed' for p in line.payments):
+                lines_w_payments.append(line)
+        if lines_w_payments:
+            names = ', '.join(
+                m.rec_name for m in lines_w_payments[:5])
+            if len(lines_w_payments) > 5:
+                names += '...'
+            key = Warning.format('reschedule_payments', lines_w_payments)
+            if Warning.check(key):
+                raise RescheduleLineWarning(
+                    key, gettext(
+                        'account_payment.msg_move_line_reschedule_payments',
+                        lines=names))
+        return super().do_reschedule(action)
+
+
+class MoveLineDelegate(metaclass=PoolMeta):
+    __name__ = 'account.move.line.delegate'
+
+    def do_delegate(self, action):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        lines_w_payments = []
+        for line in self.records:
+            if any(p.state != 'failed' for p in line.payments):
+                lines_w_payments.append(line)
+        if lines_w_payments:
+            names = ', '.join(
+                m.rec_name for m in lines_w_payments[:5])
+            if len(lines_w_payments) > 5:
+                names += '...'
+            key = Warning.format('delegate_payments', lines_w_payments)
+            if Warning.check(key):
+                raise DelegateLineWarning(
+                    key, gettext(
+                        'account_payment.msg_move_line_delegate_payments',
+                        lines=names))
+        return super().do_delegate(action)
+
+
 class Invoice(metaclass=PoolMeta):
     __name__ = 'account.invoice'
 
@@ -447,7 +603,6 @@ class Invoice(metaclass=PoolMeta):
             'invisible': Eval('type') != 'in',
             'readonly': Eval('state') != 'draft',
             },
-        depends=['type', 'state'],
         help="Check if the invoice is paid by direct debit.")
 
     @classmethod
@@ -472,25 +627,148 @@ class Invoice(metaclass=PoolMeta):
         Date = pool.get('ir.date')
         context = Transaction().context
 
-        today = Date.today()
-
         amounts = super(Invoice, cls).get_amount_to_pay(invoices, name)
 
         if context.get('with_payment', True):
-            for invoice in invoices:
-                for line in invoice.lines_to_pay:
-                    if line.reconciliation:
-                        continue
-                    if (name == 'amount_to_pay_today'
-                            and line.maturity_date
-                            and line.maturity_date > today):
-                        continue
-                    payment_amount = Decimal(0)
-                    for payment in line.payments:
-                        if payment.state != 'failed':
+            for company, c_invoices in groupby(
+                    invoices, key=lambda i: i.company):
+                with Transaction().set_context(company=company.id):
+                    today = Date.today()
+                for invoice in c_invoices:
+                    for line in invoice.lines_to_pay:
+                        if line.reconciliation:
+                            continue
+                        if (name == 'amount_to_pay_today'
+                                and line.maturity_date
+                                and line.maturity_date > today):
+                            continue
+                        payment_amount = Decimal(0)
+                        for payment in line.payments:
                             with Transaction().set_context(date=payment.date):
                                 payment_amount += Currency.compute(
-                                    payment.currency, payment.amount,
+                                    payment.currency, payment.amount_line_paid,
                                     invoice.currency)
-                    amounts[invoice.id] -= payment_amount
+                        amounts[invoice.id] -= payment_amount
         return amounts
+
+
+class Statement(metaclass=PoolMeta):
+    __name__ = 'account.statement'
+
+    @classmethod
+    def create_move(cls, statements):
+        moves = super().create_move(statements)
+        cls._process_payments(moves)
+        return moves
+
+    @classmethod
+    def _process_payments(cls, moves):
+        pool = Pool()
+        Payment = pool.get('account.payment')
+
+        to_success = defaultdict(set)
+        to_fail = defaultdict(set)
+        for move, statement, lines in moves:
+            for line in lines:
+                for kind, payments in line.payments():
+                    if (kind == 'receivable') == (line.amount >= 0):
+                        to_success[line.date].update(payments)
+                    else:
+                        to_fail[line.date].update(payments)
+        # The failing should be done last because success is usually not a
+        # definitive state
+        if to_success:
+            for date, payments in to_success.items():
+                with Transaction().set_context(clearing_date=date):
+                    Payment.succeed(Payment.browse(payments))
+        if to_fail:
+            for date, payments in to_fail.items():
+                with Transaction().set_context(clearing_date=date):
+                    Payment.fail(Payment.browse(payments))
+        if to_success or to_fail:
+            payments = set.union(*to_success.values(), *to_fail.values())
+        else:
+            payments = []
+        return list(payments)
+
+
+class StatementLine(metaclass=PoolMeta):
+    __name__ = 'account.statement.line'
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.related_to.domain['account.payment'] = [
+            ('company', '=', Eval('company', -1)),
+            If(Bool(Eval('party')),
+                ('party', '=', Eval('party')),
+                ()),
+            ('state', 'in', ['processing', 'succeeded', 'failed']),
+            ('currency', '=', Eval('currency', -1)),
+            ('kind', '=',
+                If(Eval('amount', 0) > 0, 'receivable',
+                    If(Eval('amount', 0) < 0, 'payable', ''))),
+            ]
+        cls.related_to.search_order['account.payment'] = [
+            ('amount', 'ASC'),
+            ('state', 'ASC'),
+            ]
+        cls.related_to.search_context.update({
+                'amount_order': Eval('amount', 0),
+                })
+
+    @classmethod
+    def _get_relations(cls):
+        return super()._get_relations() + ['account.payment']
+
+    @property
+    @fields.depends('related_to')
+    def payment(self):
+        pool = Pool()
+        Payment = pool.get('account.payment')
+        related_to = getattr(self, 'related_to', None)
+        if isinstance(related_to, Payment) and related_to.id >= 0:
+            return related_to
+
+    @payment.setter
+    def payment(self, value):
+        self.related_to = value
+
+    @fields.depends(
+        'party', 'statement', '_parent_statement.journal',
+        methods=['payment'])
+    def on_change_related_to(self):
+        super().on_change_related_to()
+        if self.payment:
+            if not self.party:
+                self.party = self.payment.party
+            if self.payment.line:
+                self.account = self.payment.line.account
+
+    def payments(self):
+        "Yield payments per kind"
+        if self.payment:
+            yield self.payment.kind, [self.payment]
+
+
+class Dunning(metaclass=PoolMeta):
+    __name__ = 'account.dunning'
+
+    def get_active(self, name):
+        return super().get_active(name) and self.line.payment_amount > 0
+
+    @classmethod
+    def search_active(cls, name, clause):
+        if tuple(clause[1:]) in [('=', True), ('!=', False)]:
+            domain = ('line.payment_amount', '>', 0)
+        elif tuple(clause[1:]) in [('=', False), ('!=', True)]:
+            domain = ('line.payment_amount', '<=', 0)
+        else:
+            domain = []
+        return [super().search_active(name, clause), domain]
+
+    @classmethod
+    def _overdue_line_domain(cls, date):
+        return [super()._overdue_line_domain(date),
+            ('payment_amount', '>', 0),
+            ]
